
/*
 *
 * nanotop 2.1.1
 *
 * Copyright (C) 2002 Michael Bruestle (michael.bruestle@wave-solutions.com)
 * Portions Copyright (C) 2007 Kim Jay Rogozinsky (Kim.Rogozinsky@gmail.com)
 *
 * This (2.x.x) is a quite complete re-write based upon the "nanotop" 
 * applet, ala Michael Bruestle.  The newest nanotop at the time 
 * (Jan 2007), 1.4.5d, was used as the baseline.
 *
 * We keep the original notice, as follows:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation.
 *
 */


/*
 * 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 * 12345678901234567890123456789012345678901234567890123456789012345678901234567890
 * CPU [==++--    ][==++--    ]  1234.5KB FREE  1234.5KB RX  1234.5KB TX  12345 ER
 */


#include "nanotop.h"


/*
 * Executable enty point.
 */

int main (int argc, char *argv[])
{

	/* [b]its or bytes */
	char optScale = 'B';

	/* [f]ree or used memory */
	char optMemTitle[5] = { 'F', 'R', 'E', 'E', 0 };

	/* [s]ingleline or continuous */
	char optControlChar = '\n';

	/* [k]ilo is 1000 not 1024 */
	int optKilo = 1024;

	/* [i]nterface */
	char optInterface[10] = { 'e', 't', 'h', '0', 0 };

	/* [d]elay between updates */
	long optDelay = 1;

	/* getopt */
	int locC;

	/* general use */
	int locI;

	/* tcgetattr, tcsetattr */
	struct termios locOldTTY;
	struct termios locNewTTY;

	/* main loop */
	int locExtractMem;
	int locExtractCPU0;
	int locExtractCPU1;
	int locExtractNet;
	long long locWork;
	long long locTime0;
	long long locTime1;
	long long locCPU0[4];
	long long locCPU1[4];
	long long locCPUOld0[4];
	long long locCPUOld1[4];
	long long locMEMTotal[1];
	long long locMEMFree[1];
	long long locNET[4];
	long long locNETOld[4];
	int locHaveCPU0;
	int locHaveCPU1;
	int locHaveMem;
	int locHaveNet;
	long locCountNet;
	long long locMEMValue = 0;
	char locDispCPU0[13];
	char locDispCPU1[13];
	char locDispMem[9];
	char locDispRx[9];
	char locDispTx[9];
	char locDispErr[6];
	fd_set locFDSet;
	struct timeval locTimeVal;

	/*                            */
	/* parse command line options */
	/*                            */

	while ((locC = getopt (argc, argv, "hbuoki:s:")) != -1)
	{
		switch (locC)
		{
			case 'h':
				/* [h]elp */
				print_help();
				exit(0);
			case 'b':
				/* [b]its or bytes */
				/* applies to network throughputs but not to memory */
				optScale = 'b';
				break;
			case 'u':
				/* [u]sed or free */
				strcpy (optMemTitle, "USED");
				break;
			case 'o':
				/* [o]ne line */
				optControlChar = '\r';
				break;
			case 'k':
				/* [k]ilo 1000 or 1024 */
				/* applies to network throughputs and to memory */
				optKilo = 1000;
				break;
			case 'i':
				/* [i]nterface */
				if (strlen(optarg) > 8)
				{
					printf("Err: [i]nterface\n");
					print_help_stop();
				}
				strcpy (optInterface, optarg);
				break;
			case 's':
				/* [s]econds between updates */
				optDelay = atol(optarg);
				if (optDelay < 1)
				{
					printf("Err: [s]econds\n");
					print_help_stop();
				}
				break;
			default:
				print_help_stop();
		} /* switch */
	} /* while */

	/*                              */
	/* terminal (stdin) preparation */
	/*                              */

	/* grab a copy of the current terminal settings */

	if ((tcgetattr (STDIN_FILENO, &locOldTTY)) != 0) exit(1);

	/* The new settings must be a copy of the current settings with legal alterations */
	/* Setup stdin for non-canonical read */

	locNewTTY = locOldTTY;

	locNewTTY.c_lflag &= ~(ICANON | ECHO);

	/* not sure if we need these, some use them and some don't for this situation.
	locNewTTY.c_cc[VMIN] = 1;
	locNewTTY.c_cc[VTIME] = 0;
	*/

	/* Apply our terminal settings.  We really don't care if we are successful ... it won't exit 'till we hit Enter */

	tcsetattr (STDIN_FILENO, TCSANOW, &locNewTTY);

	/* not sure if we need this, some use them and some don't for this situation.
	setbuf(stdin, NULL);
	*/

	/* display significant (non-obvious) options */

	printf ("interface: %s\n", optInterface);
	printf ("legend: =system +nice -user\n");

	/*                 */
	/* Main Loop Begin */
	/*                 */

	/* from here on we need it with a colon at the end */

	strcat (optInterface, ":");

	/* initialize variables that need to be */

	locHaveCPU0 = 0;
	locHaveCPU1 = 0;
	locCountNet = 0;

	do
	{

		/* Extract information from various files */

		locExtractCPU0 = extract ("/proc/stat", "cpu0", locCPU0, 1, 2, 3, 4, 0); /* User, Nice, System, Idle */

		/* Should not need this - even on the single processor systems I have seen, "cpu0" is reported.
		if (locExtractCPU0 != 0)
		{
			locExtractCPU0 = extract ("/proc/stat", "cpu", locCPU0, 1, 2, 3, 4, 0);
		}
		*/

		locExtractCPU1 = extract ("/proc/stat", "cpu1", locCPU1, 1, 2, 3, 4, 0); /* User, Nice, System, Idle */

		locExtractMem = extract ("/proc/meminfo", "MemTotal:", locMEMTotal, 1, 0);

		locExtractMem = extract ("/proc/meminfo", "MemFree:", locMEMFree, 1, 0);

		locExtractNet = extract ("/proc/net/dev", optInterface, locNET, 1, 3, 9, 11, 0); /* RxByte, RxErr, TxByte, TxErr */

		/* Only display CPU's we have */
		/* Only copy-compare CPU if we extracted it */
		/* Only display CPU graph if we have a valid comparison */

		if (locExtractCPU0 == 0)
		{
			locTime0 = 0;
			for (locI = 0; locI < 4; locI++)
			{
				locWork = locCPU0[locI] - locCPUOld0[locI];
				locTime0 += locWork;
				locCPUOld0[locI] = locCPU0[locI];
				locCPU0[locI] = locWork;
			}
			graph (locDispCPU0, locHaveCPU0, locTime0, locCPU0[2], locCPU0[1], locCPU0[0]);
			locHaveCPU0 = 1;
		}
		else
		{
			locDispCPU0[0] = '\0';
		}

		if (locExtractCPU1 == 0)
		{
			locTime1 = 0;
			for (locI = 0; locI < 4; locI++)
			{
				locWork = locCPU1[locI] - locCPUOld1[locI];
				locTime1 += locWork;
				locCPUOld1[locI] = locCPU1[locI];
				locCPU1[locI] = locWork;
			}
			graph (locDispCPU1, locHaveCPU1, locTime1, locCPU1[2], locCPU1[1], locCPU1[0]);
			locHaveCPU1 = 1;
		}
		else
		{
			locDispCPU1[0] = '\0';
		}

		/* Only refresh memory information if we extracted it */
		/* Currently, it appears, "MemTotal:" and "MemFree:" will always be in kilobytes. */
		/* We could code for other possibilities, however, that would require more code . */
		/* We will assume for now that they are in kilobytes. */

		locHaveMem = 0;

		if (locExtractMem == 0)
		{
			locHaveMem = 1;
			if (optMemTitle[0] == 'F')
			{
				locMEMValue = locMEMFree[0];
			}
			else
			{
				locMEMValue = locMEMTotal[0] - locMEMFree[0];
			}
		}

		/* Only refresh error information if we extracted it */
		/* Only copy-compare Rx and Tx if we extracted it */
		/* Only display Rx and Tx if we have a valid comparison */

		locHaveNet = 0;

		if (locExtractNet == 0)
		{
			snprintf (locDispErr, 6, "%5llu", locNET[1] + locNET[3]);
			for (locI = 0; locI < 4; locI += 2)
			{
				locWork = locNET[locI] - locNETOld[locI];
				locNETOld[locI] = locNET[locI];
				locNET[locI] = locWork;
			}
			locCountNet++;
			if (locCountNet > 1) locHaveNet = 1;
		}
		else
		{
			locDispErr[0] = '\0';
		}

		/*
		* Create an output like this:
		*
		* CPU [==++--    ][==++--    ]  1234.5KB FREE  1234.5KB RX  1234.5KB TX  12345 ER
		*
		* If this is the first iteration then the output will look like this:
		* CPU [          ][          ]  1234.5KB FREE           RX           TX  12345 ER
		*/

		printf
			("%cCPU %s%s  %s %s  %s RX  %s TX  %s ER", 
			optControlChar, 
			locDispCPU0, 
			locDispCPU1, 
			formatk (locDispMem, locHaveMem, 'B', optKilo, locMEMValue * 1024), 
			optMemTitle, 
			formatk (locDispRx, locHaveNet, optScale, optKilo, locNET[0] / optDelay), 
			formatk (locDispTx, locHaveNet, optScale, optKilo, locNET[2] / optDelay), 
			locDispErr);

		/* Make terminal update happen asap */

		fflush (stdout);

		/* perform efficient wait for keystroke */

		FD_ZERO(&locFDSet);
		FD_SET(STDIN_FILENO, &locFDSet);

		locTimeVal.tv_usec = 0;
		locTimeVal.tv_sec = optDelay;

		locI = select(STDIN_FILENO + 1, &locFDSet, NULL, NULL, &locTimeVal);

		/* exit if keystroke detected or select errored */

		if (locI != 0)
		{
			/* remove one keystroke from stdin, only if required */
			if (locI != -1) getchar();
			break;
		}

	} while (1);

	/*               */
	/* Main Loop End */
	/*               */

	/* Restore initial terminal settings */
	tcsetattr (STDIN_FILENO, TCSANOW, &locOldTTY);

	/* Clean up line and perhaps give message */
	printf ("\nDone\n");

	/* program terminates returning 0 (normal termination) */
	return 0;
}

/*
 * FormatK.
 */

/* Format a 64-bit integer value for presentation with "scaling"             */
/* If pasValid is 0, just return 8 spaces (like "        ")                  */
/* Present as bits or bytes                                                  */
/* Present as 1000 bytes or 1024 bytes per "kilo"                            */
/* Return a string of precisely 8 characters (like "1023.0kB" or "   1.2kB") */
/* The passed buffer *must be* (at least) 9 bytes long.                      */

char *formatk (char *pasBuffer, int pasValid, char pasScale, int pasKilo, long long pasAmount)
{
	char locPrefixes[7] = { ' ', 'k', 'M', 'G', 'T', 'P', 'E' };
	long long locDivisor;
	int locIndex;
	double locAmount;

	if (pasValid != 0)
	{

		if (pasKilo == 1024)
		{
			locPrefixes[1] = 'K';
			/*
			 * instead of locDivisor = 1152921504606846976, can also use:
			 *	exp2l(60) or powl(2, 60) or atoll("1152921504606846976") or
			 *	locDivisor = 1099511627776 -> now it's 2^40
			 *	locDivisor = locDivisor * 1048576 -> now it's 2^60
			 */
			locDivisor = 1152921504606846976;
		}
		else
		{
			/*
			 * instead of locDivisor = 1000000000000000000, can also use:
			 *	exp10l(18) or powl(10, 18) or atoll("1000000000000000000") or
			 *	locDivisor = 1000000000000 -> now it's 10^12
			 *	locDivisor = locDivisor * 1000000 -> now it's 10^18
			 */
			locDivisor = 1000000000000000000;
		}

		if (pasScale == 'b') pasAmount *= 8;

		locIndex = 6;
		do
		{
			locAmount = pasAmount / locDivisor;
			if (locAmount >= 1) break;
			if (locIndex == 0) break;
			locDivisor /= pasKilo;
			locIndex--;
		} while (1);

		snprintf (pasBuffer, 9, "%6.1f%c%c", locAmount, locPrefixes[locIndex], pasScale);

	}
	else
	{
		strcpy (pasBuffer, "        ");
	}

	return pasBuffer;
}

/*
 * Graph.
 */

/* Generate something like this:  "[==++--    ]".                                     */
/* When pasTime = 0, that particular CPU doesn't exist - return with an empty string. */
/* When pasScan = 0, the graph (albeit empty) will be returned.                       */
/* The passed buffer *must be* (at least) 13 bytes long.                              */

char * graph (char *pasBuffer, int pasValid, long long pasTime, long long pasSys, long long pasNic, long long pasUsr)
{
	int locFre = 10, locUsr = 0, locNic = 0, locSys = 0;

	if ((pasValid != 0) && (pasTime != 0))
	{
		/* turn into percentages */
		pasSys = (pasSys * 100) / pasTime;
		pasNic = (pasNic * 100) / pasTime;
		pasUsr = (pasUsr * 100) / pasTime;

		/* round percentages to tenths; 6% will become one tenth, 16% will become two tenths */
		locSys = min (locFre, (pasSys + 4) / 10);
		locFre -= locSys;
		locNic = min (locFre, (pasNic + 4) / 10);
		locFre -= locNic;
		locUsr = min (locFre, (pasUsr + 4) / 10);
		locFre -= locUsr;
	}

	snprintf (pasBuffer, 13, "[%.*s%.*s%.*s%.*s]", 
		locSys, 
		"==========", 
		locNic, 
		"++++++++++", 
		locUsr, 
		"----------", 
		locFre, 
		"          ");

	return pasBuffer;
}

/*
 * Extract fields from file.
 */

/* Return:  -1 = failed open file or did not find key on any lines. */
/*           0 = all fields found and loaded.                       */
/*        else = first field which was not found.                   */

int extract (char *pasFileName, char *pasKey, long long *pasArray, ...)
{
	int locReturn = -1;
	FILE *locFile;
	char locBuffer[256];
	char *locP1;
	va_list locArgP;
	char *locP2;
	int locSelect = 1;
	int locData = 0;
	char locSave;

	if ((locFile = fopen (pasFileName, "r")) != NULL)
	{
		while (fgets (locBuffer, sizeof (locBuffer), locFile) != NULL)
		{
			/* Find the first non white space character. */
			/* This we consider the start of the string to compare with pasKey. */
			locP1 = firstnonwhite(locBuffer);
			if (strncmp (locP1, pasKey, strlen(pasKey)) == 0)
			{
				va_start (locArgP, pasArray);
				/* text to be processed starts after key */
				locP1 += strlen(pasKey);
				/* get first required ordinal field */
				locReturn = va_arg (locArgP, int);
				/* process fields */
				while (locReturn != 0)
				{
					locP2 = firstnonwhite(locP1);
					if ((locP1 = firstwhite(locP2)) == locP2) break;
					/* is this the ordinal we are waiting for ? */
					if (locSelect == locReturn)
					{
						/* ensure the first whitespace is a null character - required for atoll */
						locSave = *locP1;
						*locP1 = '\0';
						/* convert text into long long */
						pasArray[locData] = atoll(locP2);
						/* restore original character */
						*locP1 = locSave;
						/* select next variable to store conversion in */
						locData++;
						/* get next required ordinal field */
						locReturn = va_arg (locArgP, int);
					} /* if */
					/* increment found field count */
					locSelect++;
				} /* while */
				/* must clean up when done with va_start-va_arg */
				va_end (locArgP);
				/* the appropriate line has been found and processed; we are done reading lines */
				break;
			} /*if */
		} /* while */
		/* must clean up when done with fopen */
		fclose (locFile);
	} /* if */

	return locReturn;
}

/*
 * Find first white space.
 */

char *firstwhite (char *pasPointer)
{
	/* advance while not whitespace */

	while ((*pasPointer != '\0') && (isspace(*pasPointer) == 0))
	{
		pasPointer++;
	}

	return pasPointer;
}

/*
 * Find first non-white space.
 */

char *firstnonwhite (char *pasPointer)
{
	/* advance while whitespace */

	while ((*pasPointer != '\0') && (isspace(*pasPointer) != 0))
	{
		pasPointer++;
	}

	return pasPointer;
}

/*
 * Print help and terminate in error.
 */

void print_help_stop (void)
{
	print_help();
	exit(1);
}

/*
 * Print help and return.
 */

void print_help (void)
{
	printf("nanotop %s options:\n", VERSION);
	printf("%s\n", "-h      [h]elp");
	printf("%s\n", "-b      [b]its or bytes");
	printf("%s\n", "-u      [u]sed or free");
	printf("%s\n", "-o      [o]ne line");
	printf("%s\n", "-k      [k]ilo 1000 or 1024");
	printf("%s\n", "-i eth0 [i]nterface");
	printf("%s\n", "-s 1    [s]econds between updates");
	return;
}
